import dataclasses
from typing import Generic, TypeVar, Optional, Tuple, Any, Type
from collections.abc import Callable

from autograder_platform.Executors.Environment import ImplEnvironment
from autograder_platform.StudentSubmission.AbstractStudentSubmission import AbstractStudentSubmission
from autograder_platform.StudentSubmission.ISubmissionProcess import ISubmissionProcess
from autograder_platform.StudentSubmission.SubmissionProcessFactory import SubmissionProcessFactory
from autograder_platform.config import Config
from autograder_platform.config.Config import AutograderConfiguration, AutograderConfigurationSchema
from autograder_platform.config.BaseSchema import BaseSchema

LanguageConfigType = TypeVar('LanguageConfigType')


@dataclasses.dataclass
class LanguageRegistrationMetadata(Generic[LanguageConfigType]):
    name: str
    version: str

    on_submission_process_factory_registration: Callable[
        [],
        Tuple[
            Type[AbstractStudentSubmission[Any]],
            Type[ISubmissionProcess],
            Optional[Type[ImplEnvironment]],
            Optional[Callable[[ImplEnvironment, AutograderConfiguration], None]]
        ],
    ]

    on_config_registration: Optional[Callable[[], BaseSchema[LanguageConfigType]]] = None

def register_language(metadata: LanguageRegistrationMetadata):
    if metadata.on_config_registration is not None:
        AutograderConfigurationSchema.register_sub_schema(metadata.name, metadata.on_config_registration())

    SubmissionProcessFactory.register(*metadata.on_submission_process_factory_registration())




import abc
from typing import Generic, List, Set, TypeVar, Dict, Type

from autograder_platform.StudentSubmission.ITransformer import ITransformer
from autograder_platform.StudentSubmission.common import ValidationError, ValidationHook

from autograder_platform.StudentSubmission.AbstractValidator import AbstractValidator
from autograder_platform.StudentSubmission.GenericValidators import SubmissionPathValidator

T = TypeVar("T")

# for some reason this has to be TBuilder??
Builder = TypeVar("Builder", bound="AbstractStudentSubmission[Any]")


class AbstractStudentSubmission(abc.ABC, Generic[T]):
    """
    Description
    ===========

    This class contains the abstract student submission.
    Basically, this enables more of a plug and play architecture for different submission models.

    This model also allows a cleaner and more consistent way to implement validation of the submission.

    You can implement: ref:`AbstractValidator` for your purpose and assign it a hook.
    Then, when we reach that phase of the submission, that hook will be validated.

    Subclasses must implement ``doLoad`` and ``doBuild``.
    Subclasses must also implement ``getExecutableSubmission`` which should return the submission in a state that can be executed 
    by whatever runner is implemented.

    For example, take a look at the Python (or coming soon, the C / C++) implementation.
    """
    def __init__(self):
        self.submissionRoot: str = "."
        self.validators: Dict[ValidationHook, Set[AbstractValidator]] = {}
        self.validationErrors: List[Exception] = []
        self.submissionTransformers: List[ITransformer] = []

        # default validators
        self.addValidator(SubmissionPathValidator())


    def setSubmissionRoot(self: Builder, submissionRoot: str) -> Builder:
        """
        Description
        ---
        
        Defines the root of the submission. 

        Should be a path to the ``student_work``

        :param submissionRoot: The root of the submission.
        :returns: self
        """
        self.submissionRoot = submissionRoot
        return self

    def addValidator(self: Builder, validator: AbstractValidator) -> Builder:
        """
        Description
        ---

        Adds a validator to run at the hook defined in the validator.

        Validators are provided a reference to the submission at setup time to gather 
        information that they need to know about the student's submission.
        
        The hook to use is determined by the abstract static method ``AbstractValidator.getValidationHook()``.

        Only one validator of each type is allowed, ie: If we pass two validators of type ValidateAST, 
        only the last one added will be run.
        This enforces the single responsibility principal.

        :param validator: the validator to add subject to the above information.
        :returns: self
        """
        hook = validator.getValidationHook()

        if hook not in self.validators.keys():
            self.validators[hook] = set()

        self.validators[hook].add(validator)
        return self

    def addTransformer(self: Builder, transformer: ITransformer) -> Builder:
        self.submissionTransformers.append(transformer)

        return self

    def runManualValidationHook(self, validatorType: Type[AbstractValidator]):
        if ValidationHook.MANUAL not in self.validators.keys():
            raise RuntimeError("Request to run manual hook, but no manual hooks have been defined!")

        ran = False
        for validator in self.validators[ValidationHook.MANUAL]:
            if isinstance(validator, validatorType):
                validator.setup(self)
                validator.run()
                self.validationErrors.extend(validator.collectErrors())
                # we are only allowed to have one of each type, so once we execute, we should break
                ran = True
                break

        if not ran:
            raise RuntimeError(f"Request to run manual hook {validatorType}, but that manual hook has not been defined!")

        if self.validationErrors:
            raise ValidationError(self.validationErrors)

    def runTransformers(self, submissionText: str) -> str:
        """
        Description
        ---

        Runs all transformers in the order added.
        This functionality allows the user to mutate student submissions, BUT should be used sparely.

        This is used with the iPython bindings to remove all the magic commands so that cells can be parsed into their
        AST representation.

        Transformers operate on the entire submission text, so keep that in mind when writing.

        :param submissionText: The student's entire submission text.
        Generally, should only be one file but is dependent on the implementation.
        :returns: The transformed submission.
        """

        for transformer in self.submissionTransformers:
            submissionText = transformer.transform(submissionText)

        return submissionText

    def _validate(self, validationHook: ValidationHook):
        if validationHook not in self.validators.keys():
            return

        for validator in self.validators[validationHook]:
            validator.setup(self)
            validator.run()
            self.validationErrors.extend(validator.collectErrors())

        if self.validationErrors:
            raise ValidationError(self.validationErrors)


    @abc.abstractmethod
    def doLoad(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def doBuild(self):
        raise NotImplementedError()
    
    def load(self: Builder) -> Builder:
        """
        Description
        ---

        Runs all validators attached to the ``LOAD`` hooks.

        This method calls the overriden ``doLoad`` method.

        This is step 1 in the submission pipeline.
        With the python version, this discovers submitted files and identifies the main module.

        The flow is:
        1. Run ``PRE_LOAD`` validation hook
        2. Run overriden ``doLoad``
        3. Run ``POST_LOAD`` validation hook

        :returns: self
        """
        self._validate(ValidationHook.PRE_LOAD)

        self.doLoad()

        self._validate(ValidationHook.POST_LOAD)
        return self

    def build(self: Builder) -> Builder:
        """
        Description
        ---
        
        Runs all validators attached to the ``BUILD`` hooks.

        This method calls the overriden ``doBuild`` method.

        This is step 2 in the submission pipeline.
        With the python version, this builds all the discovered files to the AST representation.

        The flow is:
        1. Run ``PRE_BUILD`` validation hook
        2. Run overriden ``doBuild``
        3. Run ``POST_BUILD`` validation hook

        :returns: self
        """
        self._validate(ValidationHook.PRE_BUILD)

        self.doBuild()

        self._validate(ValidationHook.POST_BUILD)
        return self

    def validate(self: Builder) -> Builder:
        """
        Description
        ---

        Runs all validators attached to the ``VALIDATION`` hook.

        This method should be overriden to perform any final validation that is needed prior to getting the executable version of the submission.

        :returns: self
        """
        self._validate(ValidationHook.VALIDATION)

        return self

    @abc.abstractmethod
    def getExecutableSubmission(self) -> T:
        pass

    def getSubmissionRoot(self) -> str:
        return self.submissionRoot




from typing import Callable, List, TypedDict, Tuple, Dict


class CalledWith(TypedDict):
    """
    This dictionary type stores the parameters that the function was called with
    Stores both kwargs and args
    """
    args: Tuple[object, ...]
    kwargs: Dict[str, object]

class SingleFunctionMock:
    """
    This is a simple static mock interface that allows a single function to be mocked.
    It is also pickleable, which is why the ref:`unittest.mock.Mock` could not be used in this application
    """

    def __init__(self, name: str, sideEffect: List[object] | None = None, spy: bool = False):
        self.calledTimes: int = 0
        self.calledWith: List[CalledWith] = []
        self.mockName: str = name
        self.spyFunction: Callable = self
        self.sideEffect: List[object] | None = sideEffect
        self.spy = spy

    def setSpyFunction(self, initialFunctionName: Callable):
        self.spyFunction = initialFunctionName


    def __call__(self, *args, **kwargs):
        self.calledTimes += 1
        self.calledWith.append({"args": args, "kwargs": kwargs})

        if self.spy:
            return self.spyFunction(*args, **kwargs)

        if self.sideEffect is None:
            return None

        if (self.calledTimes - 1) >= len(self.sideEffect):
            return self.sideEffect[len(self.sideEffect) - 1]

        return self.sideEffect[self.calledTimes - 1]

    def assertCalled(self):
        if self.calledTimes == 0:
            raise AssertionError(f"Function: {self.mockName} was not called.\nExpected to be called")

    def assertNotCalled(self):
        if self.calledTimes != 0:
            raise AssertionError(f"Function: {self.mockName} was called.\nExpected not to be called.")

    def assertCalledWith(self, *args, **kwargs):
        self.assertCalled()
        for calledWith in self.calledWith:
            if calledWith["args"] == args and calledWith["kwargs"] == kwargs:
                return

        raise AssertionError(f"Function: {self.mockName} was not called with arguments: {args} {kwargs}.")

    def assertCalledTimes(self, times: int):
        if times != self.calledTimes:
            raise AssertionError(
                f"Function: {self.mockName} was called {self.calledTimes} times.\nExpected to be called {times} times")

    def assertCalledAtLeastTimes(self, times: int):
        if times > self.calledTimes:
            raise AssertionError(
                f"Function: {self.mockName} was called {self.calledTimes} times.\nExpected to be called at least {times} times")



import os
import tempfile
from typing import Callable, Generic, List, Dict, Optional, Tuple, Type, TypeVar, Union, Any

import dataclasses

ImplResults = TypeVar("ImplResults")


class Results(Generic[ImplResults]):
    class Files:
        def __init__(self, files: Optional[Dict[str, str]]):
            self.files = files

        def __getitem__(self, file: str) -> Union[str, bytes]:
            if self.files is None:
                raise AssertionError(f"Missing result data. Expected: 'files'.")
            if file not in self.files:
                raise AssertionError(f"File '{file}' was not created by the student's submission!")

            readFile: Union[str, bytes] = ""

            try:
                with open(self.files[file], 'r', encoding="utf-8") as r:
                    readFile = r.read()
            except UnicodeDecodeError:
                with open(self.files[file], 'rb') as rb:
                    readFile = rb.read()

            return readFile

    def __init__(self, stdout=None, return_val=None, file_out=None, exception=None, parameters=None,
                 impl_results=None) -> None:
        self.stdout = stdout
        self.return_val = return_val
        self.file_out = file_out
        self.exception = exception
        self.parameter = parameters
        self.impl_results = impl_results

    @property
    def stdout(self) -> List[str]:
        if self._stdout is None:
            raise AssertionError(f"No OUTPUT was created by the student's submission.\n"
                                 f"Are you missing an 'OUTPUT' statement?")
        if not self._stdout:
            raise AssertionError(f"No OUTPUT was created by the student's submission.\n"
                                 f"Are you missing an 'OUTPUT' statement?")
        return self._stdout

    @stdout.setter
    def stdout(self, value: Optional[List[str]]):
        self._stdout = value

    @property
    def return_val(self) -> Optional[object]:
        return self._return_val

    @return_val.setter
    def return_val(self, value: Optional[object]):
        self._return_val = value

    @property
    def file_out(self) -> Files:
        return self._files

    @file_out.setter
    def file_out(self, value: Optional[Dict[str, str]]):
        self._files = Results.Files(value)

    @property
    def exception(self) -> Optional[Exception]:
        return self._exception

    @exception.setter
    def exception(self, value: Optional[Exception]):
        self._exception = value

    @property
    def parameter(self) -> Tuple[Any, ...]:
        if self._parameter is None:
            raise AssertionError("No parameters were set!")

        return self._parameter

    @parameter.setter
    def parameter(self, value: Optional[Tuple[Any, ...]]):
        self._parameter = value

    @property
    def impl_results(self) -> ImplResults:
        if self._impl_results is None:
            raise AssertionError("No implementation results were set!")

        return self._impl_results

    @impl_results.setter
    def impl_results(self, value: Optional[ImplResults]):
        self._impl_results = value


ImplEnvironment = TypeVar("ImplEnvironment")


@dataclasses.dataclass
class ExecutionEnvironment(Generic[ImplEnvironment, ImplResults]):
    """
    Description
    ===========

    This class defines the execution environment for the student's submission. It controls what data
    is provided and what 'pre-run' tasks are completed (ie: creating a class instance).
    This class does not define the actual executor.
    """
    sandbox_location: str = "./sandbox"
    """The location for the sandbox folder"""
    stdin: List[str] = dataclasses.field(default_factory=list)
    """If stdin will be passed to the student's submission"""
    files: Dict[str, str] = dataclasses.field(default_factory=dict)
    """What files need to be added to the students submission. 
    The key is the file name, and the value is the file name with its relative path"""
    impl_environment: Optional[ImplEnvironment] = None
    """The implementation environment options. Can be None"""
    timeout: int = 10
    """What timeout has been defined for this run of the student's submission"""
    resultData: Optional[Results[ImplResults]] = None
    """
    This dict contains the data that was generated from the student's submission. This should not be accessed
    directly, rather, use getOrAssert method
    """

def getResults(environment: ExecutionEnvironment[ImplEnvironment, ImplResults]) -> Results[ImplResults]:
    """
    This method gets the results from the environment. 
    If they aren't populated, then an assertion error is raised.
    Results should be accessed directly from the Results object by their key name.
    :param environment: the execution environment.
    :raises AssertionError: if the results aren't populated.
    """
    if environment.resultData is None:
        raise AssertionError(
            "Results are were no populated! Student submission likely crashed before writing to results object")

    return environment.resultData


Builder = TypeVar("Builder", bound="ExecutionEnvironmentBuilder[Any, Any]")
ImplEnvironmentBuilder = TypeVar("ImplEnvironmentBuilder")


class ExecutionEnvironmentBuilder(Generic[ImplEnvironment, ImplResults]):
    """
    Description
    ===========

    This class helps build the execution environments.

    See :ref:`ExecutionEnvironment` for more information
    """

    def __init__(self):
        # windows doesn't clean out temp files by default (for compatibility reasons),
        # so we are going to be good boys and girls, and delete the *contents* of this folder at the end of a run.
        # however, as it is a different folder each time, we are going to silently fail.
        self.environment = ExecutionEnvironment[ImplEnvironment, ImplResults]()
        self.dataRoot = "."

    def setDataRoot(self: Builder, dataRoot: str) -> Builder:
        """
        Description
        ---
        This function sets the data root for the execution environment, 
        meaning what prefix should be added to all the files supplied to executor

        IE: if we had a file at ``tests/data/public/file.txt``, data root should be set to 
        ``tests/data`` or ``tests/data/public``

        :param dataRoot: the data root to use.
        """
        self.dataRoot = dataRoot

        return self

    def setStdin(self: Builder, stdin: Union[List[str], str]) -> Builder:
        """
        Description
        ---
        This function sets the STDIN to supply to the student's submission.

        This should be set to the number of input lines that the submission should use.

        If stdin is supplied as a string, it will be turned into a list seperated by newlines.

        :param stdin: Either a list of input strings or a string seperated by newlines (``\n``).
        """
        if isinstance(stdin, str):
            stdin = stdin.splitlines()

        self.environment.stdin = stdin

        return self

    def addFile(self: Builder, fileSrc: str, fileDest: str) -> Builder:
        """
        Description
        ---
        This function adds a file to be pulled into the environment.

        :param fileSrc: The path to the file, relative to the specified data root. 
        IE: if we had a file at ``tests/data/public/file.txt``, and data root was set to ``tests/data``, 
        then ``fileSrc`` should be ``./public/file.txt``.
        :param fileDest: The path relative to ``SANDBOX_LOCATION`` that the file should be dropped at.
        """
        if fileSrc[0:2] == "./":
            fileSrc = fileSrc[2:]

        fileSrc = os.path.join(self.dataRoot, fileSrc)

        self.environment.files[fileSrc] = fileDest

        return self

    def setTimeout(self: Builder, timeout: int) -> Builder:
        """
        Description
        ---
        This function sets the timeout to kill the student's submission in if it doesn't end before that.

        The timeout must be integer greater than 1.

        :param timeout: The timeout to use.
        """

        self.environment.timeout = timeout

        return self

    def setImplEnvironment(self: Builder, implEnvironmentBuilder: Type[ImplEnvironmentBuilder],
                           builder: Callable[[ImplEnvironmentBuilder], ImplEnvironment]) -> Builder:

        if self.environment.impl_environment is not None:
            raise EnvironmentError("ImplEnvironment has already been defined! Should be None!")

        self.environment.impl_environment = builder(implEnvironmentBuilder())

        return self

    def _setAndResolveSandbox(self):
        tempLocation = tempfile.mkdtemp(prefix="autograder_")

        self.environment.sandbox_location = tempLocation

        for src, dest in self.environment.files.items():
            self.environment.files[src] = os.path.join(self.environment.sandbox_location, dest)

    @staticmethod
    def _validate(environment: ExecutionEnvironment):
        # For now this only validating that the files actually exist

        for src in environment.files.keys():
            if not os.path.exists(src):
                raise EnvironmentError(f"File {src} does not exist or is not accessible!")

        if not isinstance(environment.timeout, int):
            raise AttributeError(f"Timeout MUST be an integer. Was {type(environment.timeout).__qualname__}")

        if environment.timeout < 1:
            raise AttributeError(f"Timeout MUST be greater than 1. Was {environment.timeout}")

        # TODO - Validate requested features

    def build(self) -> ExecutionEnvironment[ImplEnvironment, ImplResults]:
        """
        Description
        ---
        This function validates that the execution environment is valid and then returns the environment.

        :returns: The build environment
        """

        self._setAndResolveSandbox()

        self._validate(self.environment)

        return self.environment



import abc
from typing import List

from .common import ValidationHook

class AbstractValidator(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def getValidationHook() -> ValidationHook:
        pass

    def __init__(self):
        self.errors: List[Exception] = []

    @abc.abstractmethod
    # this should be typed, but its cross-dependency issue
    def setup(self, studentSubmission):
        pass

    @abc.abstractmethod
    def run(self):
        pass

    def addError(self, error: Exception):
        self.errors.append(error)

    def collectErrors(self) -> List[Exception]:
        return self.errors



import shutil
import os
import sys
from typing import Dict

from autograder_platform.Executors.Environment import ExecutionEnvironment

from autograder_platform.StudentSubmission.SubmissionProcessFactory import SubmissionProcessFactory
from autograder_platform.Tasks.TaskRunner import TaskRunner
from autograder_platform.config.Config import AutograderConfigurationProvider, AutograderConfiguration

# For typing only
from autograder_platform.StudentSubmission.ISubmissionProcess import ISubmissionProcess


class Executor:
    @staticmethod
    def _copyFiles(files: Dict[str, str]):
        for src, dest in files.items():
            try:
                os.makedirs(os.path.dirname(dest), exist_ok=True)
                shutil.copy(src, dest)
            except OSError as ex:  # pragma: no coverage
                raise EnvironmentError(f"Failed to move file '{src}' to '{dest}'. Error is: {ex}")  # pragma: no coverage

    @classmethod
    def setup(cls, environment: ExecutionEnvironment, runner: TaskRunner, autograderConfig: AutograderConfiguration) -> ISubmissionProcess:
        if not os.path.exists(environment.sandbox_location):
            try:
                os.mkdir(environment.sandbox_location)
            except OSError as ex:  # pragma: no coverage
                raise EnvironmentError(f"Failed to create sandbox for test run. Error is: {ex}")  # pragma: no coverage

        # TODO Logging

        process = SubmissionProcessFactory.createProcess(environment, runner, autograderConfig)

        if environment.files:
            Executor._copyFiles(environment.files)

        return process
        
    @classmethod
    def execute(cls, environment: ExecutionEnvironment, runner: TaskRunner, raiseExceptions: bool = True) -> None:
        submissionProcess: ISubmissionProcess = cls.setup(environment, runner, AutograderConfigurationProvider.get())

        submissionProcess.run()

        cls.postRun(environment, submissionProcess, raiseExceptions)

    @classmethod
    def postRun(cls, environment: ExecutionEnvironment, 
                submissionProcess: ISubmissionProcess, raiseExceptions: bool) -> None:

        submissionProcess.cleanup()

        submissionProcess.populateResults(environment)

        if raiseExceptions:
            # Moving this into the actual submission process allows for each process type to
            # handle their exceptions differently
            submissionProcess.processAndRaiseExceptions(environment)


    @classmethod
    def cleanup(cls, environment: ExecutionEnvironment):
        if os.path.exists(environment.sandbox_location):
            try:
                shutil.rmtree(environment.sandbox_location)
            except OSError:  # pragma: no coverage
                pass



from typing import Callable, Final, List, Optional, Tuple
from autograder_platform.Tasks.common import TaskStatus, FailedToLoadSuppliers, AttemptToGetInvalidResults


class Task:
    def __init__(self, taskName: str, step: Callable[..., object], inputs: List[Callable[[], object]]):
        self.taskName: Final[str] = taskName
        self.step: Final[Callable[..., object]] = step
        self.inputs: Final[List[Callable[[], object]]] = inputs
        self.result: object = None
        self.status: TaskStatus = TaskStatus.NOT_STARTED
        self.error: Optional[Exception] = None

    def doTask(self):
        # TODO logging

        self.status = TaskStatus.RUNNING

        try:
            inputs: Tuple[object, ...] = tuple([getInput() for getInput in self.inputs])
        except Exception as ex:
            # TODO logging
            self.status = TaskStatus.ERROR
            self.error = FailedToLoadSuppliers(ex)
            return

        try:
            self.result = self.step(*inputs)
        except Exception as ex:
            # TODO logging
            self.status = TaskStatus.ERROR
            self.error = ex
            return

        self.status = TaskStatus.COMPLETE
        # TODO logging

    def getResult(self) -> object:
        if self.error or self.status != TaskStatus.COMPLETE:
            raise AttemptToGetInvalidResults()

        return self.result

    def getStatus(self) -> TaskStatus:
        return self.status

    def getName(self) -> str:
        return self.taskName

    def getError(self) -> Optional[Exception]:
        return self.error



import math
import re
import unittest
from typing import Optional, List, Any


class Assertions(unittest.TestCase):
    """
    This class contains the base assertions for the autograder platform. It overrides the one in the base TestCase class

    The primary differentiation factor of this is that it formats the outputs in a nicer way for both gradescope and the
    local autograder
    """
    def __init__(self, testResults):
        super().__init__(testResults)
        self.addTypeEqualityFunc(str, self.assertMultiLineEqual)
        # self.addTypeEqualityFunc(dict, self.assertDictEqual)
        self.addTypeEqualityFunc(list, self.assertListEqual)
        self.addTypeEqualityFunc(tuple, self.assertTupleEqual)

    @staticmethod
    def _stripOutput(outputLine: str) -> str:
        if not isinstance(outputLine, str):
            raise AssertionError(f"Expected a string. Got {type(outputLine).__qualname__}")

        if "output " in outputLine.lower():
            outputLine = outputLine[7:]

        outputLine = outputLine.strip()

        return outputLine

    @staticmethod
    def _convertStringToList(outputLine: str) -> list[str]:
        if not isinstance(outputLine, str):
            raise AssertionError(f"Expected a string. Got {type(outputLine).__qualname__}")

        outputLine = Assertions._stripOutput(outputLine)

        if "[" in outputLine:
            outputLine = outputLine[1:]
        if "]" in outputLine:
            outputLine = outputLine[:-1]
        outputLine = outputLine.strip()

        parsedList: list[str] = outputLine.split(",")
        charsToRemove = re.compile("[\"']")
        parsedList = [el.strip() for el in parsedList]
        parsedList = [re.sub(charsToRemove, "", el) for el in parsedList]
        return parsedList

    @staticmethod
    def _raiseFailure(shortDescription: str, expectedObject: object, actualObject: object, msg: Optional[str]):
        errorMsg = f"Incorrect {shortDescription}.\n" + \
                   f"Expected {shortDescription}: {expectedObject}\n" + \
                   f"Your {shortDescription}    : {actualObject}"
        if msg:
            errorMsg += "\n\n" + str(msg)

        raise AssertionError(errorMsg)

    @staticmethod
    def _convertIterableFromString(expected, actual):
        for i in range(len(expected)):
            parsedActual = object
            expectedType = type(expected[i])
            try:
                if isinstance(actual[i], type(expected[i])):
                    parsedActual = actual[i]
                elif expected[i] is None:
                    if actual[i] == "None":
                        parsedActual = None
                elif isinstance(expected[i], bool):
                    parsedActual = True if actual[i] == "True" else False
                else:
                    parsedActual = expectedType(actual[i])
            except Exception:
                raise AssertionError(f"Failed to parse {expectedType.__qualname__} from {actual[i]}")

            actual[i] = parsedActual

        return actual

    def _assertIterableEqual(self, expected, actual, msg: Optional[str] = None):
        for i in range(len(expected)):
            if expected[i] != actual[i]:
                self._raiseFailure("output", expected[i], actual[i], msg)

    @staticmethod
    def findPrecision(x: float):
        """
        This function is stolen from stack overflow verbatim - it computes the precision of a float
        """
        max_digits = 14
        int_part = int(abs(x))
        magnitude = 1 if int_part == 0 else int(math.log10(int_part)) + 1
        if magnitude >= max_digits:
            return magnitude
        frac_part = abs(x) - int_part
        multiplier = 10 ** (max_digits - magnitude)
        frac_digits = multiplier + int(multiplier * frac_part + 0.5)
        while frac_digits % 10 == 0:
            frac_digits /= 10
        scale = int(math.log10(frac_digits))
        return magnitude + scale

    def assertMultiLineEqual(self, expected: str, actual: str, msg: Optional[str] = None) -> None:
        if not isinstance(expected, str):
            raise AttributeError(f"Expected must be string. Actually is {type(expected).__qualname__}")
        if not isinstance(actual, str):
            raise AssertionError(f"Expected a string. Got {type(actual).__qualname__}")

        if expected != actual:
            self._raiseFailure("output", expected, actual, msg)

    def _assertListPreCheck(self, expected: list[any], actual: list[object] | str, msg: Optional[str] = None):
        if not isinstance(expected, list):
            raise AttributeError(f"Expected must be a list. Actually is {type(expected).__qualname__}")
        if isinstance(actual, str):
            actual = self._convertStringToList(actual)

        if not isinstance(actual, list):
            raise AssertionError(f"Expected a list. Got {type(actual).__qualname__}")

        if len(expected) != len(actual):
            self._raiseFailure("number of elements", len(expected), len(actual), msg)

        return self._convertIterableFromString(expected, actual)

    def assertListEqual(self, expected: list[any], actual: list[object] | str, msg: Optional[str] = None) -> None:
        actual = self._assertListPreCheck(expected, actual, msg)
        self._assertIterableEqual(expected, actual, msg)

    def assertListAlmostEqual(self, expected: list[any], actual: list[object] | str, allowedDelta: float,
                              msg: Optional[str] = None) -> None:
        actual = self._assertListPreCheck(expected, actual, msg)
        for i in range(len(expected)):
            self.assertAlmostEquals(expected[i], actual[i], delta=allowedDelta)

    def assertTupleEqual(self, expected: tuple[any, ...], actual: tuple[object, ...], msg: Optional[str] = None) -> None:
        if not isinstance(expected, tuple):
            raise AttributeError(f"Expected must be a tuple. Actually is {type(expected).__qualname__}")

        if not isinstance(actual, tuple):
            raise AssertionError(f"Expected a tuple. Got {type(actual).__qualname__}")

        if len(expected) != len(actual):
            self._raiseFailure("number of elements", len(expected), len(actual), msg)

        self._assertIterableEqual(expected, actual, msg)

    def assertDictEqual(self, expected: dict[any, object], actual: dict[object, object], msg: Optional[str] = None) -> None:
        raise NotImplementedError("Use base assert dict equal")

    def assertAlmostEquals(self, expected: float, actual: float, places: int = ..., msg: Optional[str] = None,
                           delta: float = ...) -> None:
        if places is None:
            raise AttributeError("Use delta not places for assertAlmostEquals")

        if round(abs(expected - actual), self.findPrecision(delta)) > delta:
            self._raiseFailure(f"output (allowed delta +/- {delta})", expected, actual, msg)


    @staticmethod
    def assertCorrectNumberOfOutputLines(expected: List[Any], actual: List[Any]):
        if len(actual) == 0:
            raise AssertionError("No OUTPUT lines found. Check OUTPUT formatting.")

        if len(actual) > len(expected):
            raise AssertionError(f"Too many OUTPUT lines. Check OUTPUT formatting.\n"
                                 f"Expected number of lines: {len(expected)}\n"
                                 f"Actual number of lines  : {len(actual)}")

        if len(actual) < len(expected):
            raise AssertionError(f"Too few OUTPUT lines. Check OUTPUT formatting.\n"
                                 f"Expected number of lines: {len(expected)}\n"
                                 f"Actual number of lines  : {len(actual)}")



from typing import Dict, Generic, TypeVar
from abc import ABC, abstractmethod

T = TypeVar("T")

BaseSchemaType = TypeVar('BaseSchemaType', bound='BaseSchema[Any]')

class BaseSchema(Generic[T], ABC):
    _registered_sub_schemas: Dict[str, BaseSchemaType] = {}

    @classmethod
    def register_sub_schema(cls, language_name: str, sub_schema: BaseSchemaType):
        cls._registered_sub_schemas[language_name] = sub_schema

    @classmethod
    def deregister_sub_schema(cls, language_name: str):
        del cls._registered_sub_schemas[language_name]

    @abstractmethod
    def validate(self, data: Dict) -> Dict:
        raise NotImplementedError()

    @abstractmethod
    def build(self, data: Dict) -> T:
        raise NotImplementedError()




import enum


class TaskStatus(enum.Enum):
    NOT_STARTED = 1
    RUNNING = 2
    COMPLETE = 3
    ERROR = 4


class FailedToLoadSuppliers(Exception):
    def __init__(self, ex: Exception):
        super().__init__(f"Failed to process input suppliers.\n{ex}")

class AttemptToGetInvalidResults(Exception):
    def __init__(self):
        super().__init__("Attempt to get result from task with: Status != COMPLETE")

class TaskAlreadyExists(Exception):
    def __init__(self, taskName: str):
        super().__init__(f"Task '{taskName}' has already been defined in this task runner!")


class TaskDoesNotExist(Exception):
    def __init__(self, taskName: str):
        super().__init__(f"Task '{taskName}' does not exist!")


import os
import re
import shutil
from enum import Enum
from typing import List, Dict, Callable

from autograder_platform.cli import AutograderCLITool
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfiguration


class FilesEnum(Enum):
    PUBLIC_TEST = 0
    PRIVATE_TEST = 1
    PUBLIC_DATA = 2
    PRIVATE_DATA = 3
    STARTER_CODE = 4
    CONFIG_FILE = 5


class Build:
    IGNORE = ["__pycache__"]

    def __init__(self, config: AutograderConfiguration, sourceRoot, binRoot, version) -> None:
        self.config = config
        self.generationDirectory = os.path.join(binRoot, "generation")
        self.distDirectory = os.path.join(binRoot, "dist")
        self.binRoot = binRoot
        self.sourceDir = sourceRoot
        self.version = version

    @staticmethod
    def _discoverTestFiles(allowPrivate: bool,
                           privateTestFileRegex: re.Pattern, publicTestFileRegex: re.Pattern,
                           testDirectory: str,
                           discoveredPrivateFiles: List[str], discoveredPublicFiles: List[str]):
        """
        Description
        ---

        This function discovers the test files in the testDirectory and puts them in the correct list
        based on the regex matches.

        Note: If allowPrivate is false, then all files that match the private regex will be added to the public list

        :param allowPrivate: If private files should be treated as private
        :param privateTestFileRegex: The regex pattern used to match private test files
        :param publicTestFileRegex: The regex pattern used to match public test files
        :param discoveredPrivateFiles: The list that contains the private files to be copied
        :param discoveredPublicFiles: The list that contains the public files to be copied
        """

        # test discovery is non recursive for now
        test_files = [file for file in os.listdir(testDirectory) if os.path.isfile(os.path.join(testDirectory, file))]

        for file in test_files:
            path = os.path.join(testDirectory, file)
            # Dont need to worry about double checking, the private test will only be run once in both of these cases
            if allowPrivate and privateTestFileRegex.match(file):
                discoveredPrivateFiles.append(path)
            elif publicTestFileRegex.match(file) or privateTestFileRegex.match(file):
                discoveredPublicFiles.append(path)

    @staticmethod
    def _discoverDataFiles(allowPrivate: bool, dataFilesSource: str,
                           discoveredPrivateFiles: List[str], discoveredPublicFiles: List[str]):
        """
        Description
        ---

        This function recursively discovers the data files in the dataFilesSource.
        As opposed to the test file function, this will mark files as private if they contain 'private' anywhere in the path.

        Note: if allowPrivate is false, then all files that would otherwise be private will be added to the public list

        In the godforsaken event that some how we have a directory structure that exceeds 1000 folders, this will fail
        due to a recursion error

        :param allowPrivate: If private files should be treated as private
        :param dataFilesSource: The current search directory
        :param discoveredPrivateFiles: The list that contains the private files to be copied
        :param discoveredPublicFiles: the list that contains the public files to be copied
        """

        for file in os.listdir(dataFilesSource):
            # ignore hidden files + directories
            if file[0] == ".":
                continue

            path = os.path.join(dataFilesSource, file)

            # ignore any and all test files
            if os.path.isfile(path) and "test" in file.lower():
                continue

            if os.path.isdir(path):
                Build._discoverDataFiles(allowPrivate, path, discoveredPrivateFiles, discoveredPublicFiles)
                continue

            if allowPrivate and "private" in path.lower():
                discoveredPrivateFiles.append(path)
                continue

            discoveredPublicFiles.append(path)

    def discoverFiles(self) -> Dict[FilesEnum, List[str]]:
        """
        Description
        ---

        This function discovers all of the user defined files to copy.
        See :ref:`_discoverTestFiles` and :ref:`_discoverDataFiles` for more information on how this process works

        :returns: A map containing all the user defined files to copy
        """
        config = self.config.build

        files: Dict[FilesEnum, List[str]] = {
            FilesEnum.PUBLIC_TEST: [],
            FilesEnum.PRIVATE_TEST: [],
            FilesEnum.PUBLIC_DATA: [],
            FilesEnum.PRIVATE_DATA: [],
            FilesEnum.STARTER_CODE: [],
            FilesEnum.CONFIG_FILE: [],
        }

        self._discoverTestFiles(config.allow_private,
                                re.compile(config.private_tests_regex), re.compile(config.public_tests_regex),
                                self.config.config.test_directory,
                                files[FilesEnum.PRIVATE_TEST], files[FilesEnum.PUBLIC_TEST])

        # imo, this is not worth moving into its function atm
        if config.use_starter_code:
            # we can assume that the file exists if the config has it
            files[FilesEnum.STARTER_CODE].append(config.starter_code_source)

        if config.use_data_files:
            self._discoverDataFiles(config.allow_private, config.data_files_source,
                                    files[FilesEnum.PRIVATE_DATA], files[FilesEnum.PUBLIC_DATA])

        files[FilesEnum.CONFIG_FILE] = [os.path.join(self.sourceDir, "config.toml")]

        return files

    @staticmethod
    def copy(src, dest):
        if os.path.isdir(src):
            shutil.copytree(src, dest)
            return

        shutil.copy(src, dest)

    def createFolders(self):
        # clean build if it exists
        if os.path.exists(self.binRoot):
            try:
                shutil.rmtree(self.binRoot, ignore_errors=True)
            except OSError:  # pragma: no coverage
                print("WARN: Failed to clean bin directory")  # pragma: no coverage

        # create directories
        os.makedirs(self.generationDirectory, exist_ok=True)
        os.makedirs(self.distDirectory, exist_ok=True)

    @staticmethod
    def createSetupForGradescope(path: str, version: str):
        with open(os.path.join(path, "setup.sh"), "w") as w:
            w.write(
                "add-apt-repository ppa:deadsnakes/ppa -y\n"
                "apt update\n"
                "apt-get install python3.12 -y\n"
                "apt-get install python3.12-venv -y\n"
                "python3.12 -m venv /autograder/.venv\n"
                "source /autograder/.venv/bin/activate\n"
                f"pip install 128Autograder=={version}\n"
            )

    @staticmethod
    def createRunFileForGradescope(path: str):
        with open(os.path.join(path, "run_autograder"), "w") as w:
            w.write(
                "#!/bin/bash\n"
                "source /autograder/.venv/bin/activate\n"
                "pushd source > /dev/null || echo 'Autograder failed to open source'\n"
                "run_gradescope\n"
                "popd > /dev/null || true\n"
            )
        os.chmod(os.path.join(path, "run_autograder"), 0o777)

    @staticmethod
    def createRunForPrairieLearn(path: str):
        with open(os.path.join(path, "run_autograder"), "w") as w:
            w.write(
                "#!/bin/bash\n"
                "cd /grade > /dev/null || echo 'Autograder failed to open source'\n"
                "mkdir -p results\n"
                "echo '{}' > /grade/results/results.json\n"
                "run_prairielearn --config-file /grade/tests/config.toml || exit 0\n"
            )
        os.chmod(os.path.join(path, "run_autograder"), 0o777)

    @staticmethod
    def generateDocker(generationPath: str, platform, files: Dict[FilesEnum, List[str]], version: str,
                       setupFileGenerator: Callable[[str, str], None], runFileGenerator: Callable[[str], None]):
        generationPath = os.path.join(generationPath, "docker", platform)
        os.makedirs(generationPath, exist_ok=True)

        setupFileGenerator(generationPath, version)
        runFileGenerator(generationPath)

        for key, listOfFiles in files.items():
            if key is FilesEnum.STARTER_CODE:
                continue

            for file in listOfFiles:
                destPath = os.path.join(generationPath, file)
                os.makedirs(os.path.dirname(destPath), exist_ok=True)
                Build.copy(file, destPath)

    @staticmethod
    def generatePrairieLearn(generationPath: str, files: Dict[FilesEnum, List[str]], runFileGenerator: Callable[[str], None]):
        # autograder needs to be generated in the test folder as that is the only folder that PL will mount
        generationPath = os.path.join(generationPath, "prairielearn")
        sourcePath = os.path.join(generationPath, "tests")

        os.makedirs(generationPath, exist_ok=True)
        os.makedirs(sourcePath, exist_ok=True)

        runFileGenerator(sourcePath)

        for key, listOfFiles in files.items():
            if key is FilesEnum.STARTER_CODE:
                for file in listOfFiles:
                    destPath = os.path.join(generationPath, os.path.basename(file))
                    os.makedirs(os.path.dirname(destPath), exist_ok=True)
                    Build.copy(file, destPath)
                continue

            for file in listOfFiles:
                destPath = os.path.join(sourcePath, file)
                os.makedirs(os.path.dirname(destPath), exist_ok=True)
                Build.copy(file, destPath)



    @staticmethod
    def generateStudent(generationPath: str, files: Dict[FilesEnum, List[str]], studentWorkFolder: str):
        generationPath = os.path.join(generationPath, "student")
        os.makedirs(generationPath, exist_ok=True)

        # create student_work folder
        studentWorkFolder = os.path.join(generationPath, studentWorkFolder)
        os.makedirs(studentWorkFolder, exist_ok=True)

        for file in files[FilesEnum.PUBLIC_TEST]:
            destPath = os.path.join(generationPath, file)
            os.makedirs(os.path.dirname(destPath), exist_ok=True)
            Build.copy(file, destPath)

        for file in files[FilesEnum.PUBLIC_DATA]:
            destPath = os.path.join(generationPath, file)
            os.makedirs(os.path.dirname(destPath), exist_ok=True)
            Build.copy(file, destPath)
            # also add to student work folder
            destPath = os.path.join(studentWorkFolder, os.path.basename(file))
            Build.copy(file, destPath)

        for file in files[FilesEnum.STARTER_CODE]:
            destPath = os.path.join(studentWorkFolder, os.path.basename(file))
            Build.copy(file, destPath)

        for file in files[FilesEnum.CONFIG_FILE]:
            destPath = os.path.join(generationPath, os.path.basename(file))
            Build.copy(file, destPath)

        # create .keep so that we dont loose the file
        with open(os.path.join(studentWorkFolder, ".keep"), "w") as w:
            w.write("DO NOT WRITE YOUR CODE HERE!\nCreate a *new* file in this directory!!!")

    @staticmethod
    def createDist(distType: str, generationPath: str, distPath: str, assignmentName: str):
        generationPath = os.path.join(generationPath, distType)
        if not os.path.exists(generationPath) or not os.path.isdir(generationPath):
            raise AttributeError(f"Invalid generation path: {generationPath}")

        os.makedirs(distPath, exist_ok=True)

        assignmentName += f"-{'-'.join(distType.split('/'))}"
        distPath = os.path.join(distPath, assignmentName)

        shutil.make_archive(distPath, "zip", root_dir=generationPath)

    def build(self):
        files = self.discoverFiles()

        self.createFolders()

        if self.config.build.build_gradescope:
            self.generateDocker(self.generationDirectory, "gradescope", files, self.version,
                                self.createSetupForGradescope, self.createRunFileForGradescope)
            self.createDist("docker/gradescope", self.generationDirectory, self.distDirectory,
                            f"{self.config.semester}_{self.config.assignment_name}")
        if self.config.build.build_prairie_learn:
            # this build is a touch bigger than it needs to be, but for now I'm not super concerned about it
            self.generatePrairieLearn(self.generationDirectory,  files, self.createRunForPrairieLearn)
            self.createDist("prairielearn", self.generationDirectory, self.distDirectory,
                            f"{self.config.semester}_{self.config.assignment_name}")

        if self.config.build.build_student:
            self.generateStudent(self.generationDirectory, files,
                                 self.config.build.student_work_folder)
            self.createDist("student", self.generationDirectory, self.distDirectory,
                            f"{self.config.semester}_{self.config.assignment_name}")


class BuildAutograderCLI(AutograderCLITool):
    def __init__(self):
        super().__init__(f"Builder v{AutograderCLITool.get_version()}")

    def configure_options(self):
        self.parser.add_argument("--source", default=".", help="Autograder source root")
        self.parser.add_argument("-o", default="./bin", help="Output folder")

    def run(self) -> bool:
        self.configure_options()
        self.load_config()

        build = Build(self.config, self.arguments.source, self.arguments.o, self.get_version())

        build.build()

        return False

    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):
        pass


tool = BuildAutograderCLI().run

if __name__ == "__main__":
    res = tool()
    exit(res)



import hashlib
import json
import os
from typing import Dict, List

from autograder_utils.ResultBuilders import gradescopeResultBuilder
from autograder_utils.ResultFinalizers import gradescopeResultFinalizer
from autograder_utils.JSONTestRunner import JSONTestRunner

# CLI tools should only be able to import from the CLI part of the library
from autograder_platform.cli import AutograderCLITool
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfiguration


class GradescopeAutograderCLI(AutograderCLITool):
    def __init__(self):
        super().__init__("Gradescope")

    def gradescope_post_processing(self, autograderResults: Dict, acceptable_hash: str):
        if not os.path.exists(self.arguments.metadata_path):
            autograderResults['output'] = "Autograder run was INVALID. Please resubmit. This will not count against your submission limit"
            return

        if "tests" not in autograderResults or len(autograderResults["tests"]) == 0:
            autograderResults['output'] = "No tests were run. If you are a student seeing this message, please notify course staff."
            return

        if "score" not in autograderResults:
            autograderResults['output'] = "Autograder run was INVALID. Please resubmit. This will not count against your submission limit"
            return

        # for now, we aren't implementing any new features for this
        submissionLimit = self.config.config.submission_limit
        takeHighest = self.config.config.take_highest

        # Enforce submission limit
        submissionMetadata: Dict = {}

        if self.read_hash(self.arguments.metadata_path) != acceptable_hash:
            autograderResults['output'] = "Autograder run was INVALID. Please resubmit. This will not count against your submission limit"
            return

        with open(self.arguments.metadata_path, 'r') as submissionMetadataIn:
            submissionMetadata = json.load(submissionMetadataIn)

        previousSubmissions: List[Dict] = submissionMetadata['previous_submissions']

        # this grabs only the valid previous submissions - must have a result, and a score
        validSubmissions: List[Dict] = \
            [previousSubmissionMetadata['results']
             for previousSubmissionMetadata in previousSubmissions
             if 'results' in previousSubmissionMetadata.keys() and "score" in previousSubmissionMetadata["results"].keys()
             ]


        if self.config.config.enforce_submission_limit:
            autograderResults['output'] = f"Submission {len(validSubmissions) + 1} of {submissionLimit}.\n"
        else:
            autograderResults['output'] = ""

        validSubmissions.append(autograderResults)

        # submission limit exceeded
        if self.config.config.enforce_submission_limit and len(validSubmissions) > submissionLimit:
            autograderResults['output'] += f"Submission limit exceeded.\n" \
                                           f"Autograder has been run on your code so you can see how you did\n" \
                                           f"but, your score will be highest of your valid submissions.\n"
            validSubmissions = validSubmissions[:submissionLimit]
            # We should take the highest valid submission
            takeHighest = True

        # sorts in descending order
        validSubmissions.sort(reverse=True, key=lambda submission: submission['score'] if 'score' in submission else 0)

        if takeHighest and validSubmissions[0] != autograderResults:
            autograderResults['output'] += f"Score has been set to your highest valid score.\n"
            autograderResults['score'] = validSubmissions[0]['score']

        # ensure that negative scores arent possible
        if autograderResults['score'] < 0:
            autograderResults['output'] += f"Score has been set to a floor of 0 to ensure no negative scores.\n"
            autograderResults['score'] = 0

        max_score = self.config.config.max_score if self.config.config.allow_extra_credit else self.config.config.perfect_score

        if autograderResults["score"] > max_score:
            autograderResults["output"] += f"Score has been capped to {max_score}.\n"
            autograderResults["score"] = max_score



    def configure_options(self):  # pragma: no cover
        self.parser.add_argument("--results-location", default="/autograder/results/results.json",
                                 help="The location for the autograder JSON results")
        self.parser.add_argument("--metadata-path", default="/autograder/submission_metadata.json",
                                 help="The location for the submission metadata JSON")
        self.parser.add_argument("--submission-directory", default="/autograder/submission",
                                 help="The directory where the student's submission is located")


    def read_hash(self, metadata_path):
        if not os.path.exists(metadata_path):
            return False

        with open(metadata_path, 'rb') as rb:
            fileBytes = rb.read()
        return hashlib.md5(fileBytes, usedforsecurity=False).hexdigest()

    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):  # pragma: no cover
        if self.arguments is None:
            return

        configBuilder.setStudentSubmissionDirectory(self.arguments.submission_directory)

    def run(self) -> bool:  # pragma: no cover
        self.configure_options()

        self.load_config()

        if self.arguments is None:
            return True

        self.discover_tests()

        acceptable_hash = self.read_hash(self.arguments.metadata_path)

        with open(self.arguments.results_location, 'w') as w:
            testRunner = JSONTestRunner(visibility='visible', stream=w,
                                        result_builder=gradescopeResultBuilder,
                                        result_finalizer=gradescopeResultFinalizer,
                                        post_processor=lambda results: self.gradescope_post_processing(results, acceptable_hash))

            res = testRunner.run(self.tests)

            return not res.wasSuccessful()


tool = GradescopeAutograderCLI().run

if __name__ == "__main__":
    res = tool()

    exit(res)




import importlib
import os
from tomli import load
from typing import Dict, Generic, Optional as OptionalType, TypeVar, Any
from dataclasses import dataclass

from schema import And, Optional, Regex, Schema, SchemaError

from autograder_platform.config.common import InvalidConfigException
from autograder_platform.config.BaseSchema import BaseSchema

LanguageConfigType = TypeVar('LanguageConfigType')

@dataclass(frozen=True)
class BuildConfiguration:
    """
    Build Configuration
    ===================
    
    This class defines the build options when building new or existing autograders
    """
    use_starter_code: bool
    """Whether the starter code should be pulled into the student autograder"""
    use_data_files: bool
    """Whether datafiles should be pulled into the autograder"""
    allow_private: bool
    """Whether the private tests and datafiles should be kept private"""
    build_student: bool
    """If we should build the student autograder"""
    build_gradescope: bool
    """If we should build the gradescope autograder"""
    build_prairie_learn: bool
    """If we should build the PrairieLearn autograder"""
    data_files_source: str
    """The folder that contains the datafiles for use with the autograder"""
    starter_code_source: str
    """The path for the starter code that should be provided to the student"""
    student_work_folder: str
    """The folder that should be created for the student"""
    private_tests_regex: str
    """The pattern that should be used to identify private tests"""
    public_tests_regex: str
    """The pattern that should be used to identify public tests"""

@dataclass(frozen=True)
class BasicConfiguration:
    """
    Basic Configuration
    ===================

    This class defines the basic autograder configuration
    """
    language_to_use: str
    """The language to use"""
    student_submission_directory: str
    """The folder that the student submission is in"""
    autograder_version: str
    """The autograder version tag. Autograder will be kept at this version"""
    test_directory: str
    """The directory that the tests are in"""
    enforce_submission_limit: bool
    """Whether or not the submission limit should be enforced"""
    submission_limit: int
    """The max number of submissions that a student has"""
    take_highest: bool
    """If we should take the highest of all the valid scores"""
    allow_extra_credit: bool
    """If scores greater than ``perfect_score`` should be respected"""
    perfect_score: int
    """The max score with out extra credit that a student can get"""
    max_score: int
    """
    The max score that students can get with extra credit. 
    Points greater than this will not be honored.
    """
    submission_files: list
    """
    An optional list of expected submission files when submission files other than main/submission are submitted
    """

@dataclass(frozen=True)
class AutograderConfiguration(Generic[LanguageConfigType]):
    """
    Autograder Configuration
    ========================

    The root object for the autograder configuration.
    Comprised of sub objects. 
    See :ref:`BasicConfiguration` and :ref:`BuildConfiguration` for more information.
    """
    assignment_name: str
    """The assignment name. IE: `ConstructionSite`"""
    semester: str
    """The semester that this assignment is being offered. IE: F99 -> Fall 1999"""
    autograder_root: str
    """The autograder's root directory where the config.toml file is located"""
    config: BasicConfiguration
    """The basic settings for the autograder. See :ref:`BasicConfiguration` for options."""
    language_config: OptionalType[LanguageConfigType]
    """The language config for the autograder. See the the language's config for options."""
    build: BuildConfiguration
    """The build configuration for the autograder. See :ref:`BuildConfiguration` for options."""


class AutograderConfigurationSchema(BaseSchema[AutograderConfiguration]):
    """
    Autograder Configuration Schema
    ===============================

    This class defines the format agnostic schema required for the autograder.
    This class is able to validate and build a config. 
    Configs are expected to be provided as a dictionary; hence that agnostic nature of the schema.

    This class builds to: ref:`AutograderConfiguration` for easy typing.
    """

    @classmethod
    def validateImplSource(cls, implName: str) -> bool:
        return implName in cls._registered_sub_schemas

    def __init__(self):
        self.currentSchema: Schema = Schema(
            {
                "assignment_name": And(str, Regex(r"^(\w+-?)+$")),
                "semester": And(str, Regex(r"^(F|S|SUM)\d{2}$")),
                # TODO: need to make this use the current config file name
                Optional("autograder_root", default="."): And(os.path.exists, os.path.isdir, lambda path: any([".toml" in file for file in os.listdir(path)])),
                "config": {
                    "language_to_use": And(str, AutograderConfigurationSchema.validateImplSource),
                    Optional("student_submission_directory", default="."): And(str, os.path.exists, os.path.isdir),
                    "autograder_version": And(str, Regex(r"\d+\.\d+\.\d+")),
                    "test_directory": And(str, os.path.exists),
                    "enforce_submission_limit": bool,
                    Optional("submission_limit", default=1000): And(int, lambda x: x >= 1),
                    Optional("take_highest", default=True): bool,
                    Optional("allow_extra_credit", default=False): bool,
                    "perfect_score": And(int, lambda x: x >= 1),
                    "max_score": And(int, lambda x: x >= 1),
                    Optional("submission_files", default=[]): And(list, lambda x: len(x)>0)
                },
                "build": {
                    "use_starter_code": bool,
                    "use_data_files": bool,
                    Optional("allow_private", default=True): bool,
                    Optional("data_files_source", default=None): str,
                    Optional("starter_code_source", default=None): str,
                    "build_student": bool,
                    "build_gradescope": bool,
                    Optional("build_prairie_learn", default=False): bool,
                    Optional("student_work_folder", default="student_work"): str,
                    Optional("private_tests_regex", default=r"^test_private_?\w*\.py$"): str,
                    Optional("public_tests_regex", default=r"^test_?\w*\.py$"): str,
                },
                # allow extra top level keys to be caught so we dont have to discard them
                object: dict,
            },
            ignore_extra_keys=True, name="ConfigSchema"
        )

    def validate(self, data: Dict) -> Dict:
        """
        Description
        ---
        This method validates the provided data against the schema.

        If it is valid, then it will be returned. Otherwise, an ``InvalidConfigException`` will be raised.

        :param data: The data to validate
        :return: The data if it is able to be validated
        """
        validated = {}
        try:
            validated = self.currentSchema.validate(data)
        except SchemaError as schemaError:
            raise InvalidConfigException(str(schemaError))

        if validated["config"]['language_to_use'] in self._registered_sub_schemas:
            validated = self._registered_sub_schemas[validated["config"]['language_to_use']].validate(validated)

        impl_to_use = validated["config"]["language_to_use"]

        if impl_to_use not in validated or validated[impl_to_use] is None:
            raise InvalidConfigException(f"Missing Implementation Config for {impl_to_use}")

        if validated["build"]["use_starter_code"] and validated["build"]["starter_code_source"] is None:
            raise InvalidConfigException("Missing starter code file location")

        if validated["build"]["use_data_files"] and validated["build"]["data_files_source"] is None:
            raise InvalidConfigException("Missing directory for data files!")

        return validated

    def build(self, data: Dict) -> AutograderConfiguration[LanguageConfigType]:
        """
        Description
        ---
        This method builds the provided data into the known config format.

        In this case, it builds into the ``AutograderConfiguration`` format.
        Data should be validated before calling this method as it uses dictionary expansion to populate the config objects.

        Doing this allows us to have a strongly typed config format to be used later in the autograder.
        """

        language_config: OptionalType[LanguageConfigType] = None

        if data["config"]['language_to_use'] in self._registered_sub_schemas and data[data['config']['language_to_use']] is not None:
            language_config = self._registered_sub_schemas[data['config']['language_to_use']].build(data)
        else:
            raise InvalidConfigException("Language to use has not been registered!")

        data['language_config'] = language_config

        del data[data["config"]['language_to_use']]

        data["config"] = BasicConfiguration(**data["config"])
        data["build"] = BuildConfiguration(**data["build"])

        return AutograderConfiguration(**data)


# Using generics as PyRight and mypy are able to infer what `T` should be from the Schema
#  as it inherits from BaseSchema
T = TypeVar("T")
Builder = TypeVar("Builder", bound="AutograderConfigurationBuilder[Any]")


class AutograderConfigurationBuilder(Generic[T]):
    """
    AutograderConfigurationBuilder
    ==============================

    This class currently doesn't do much.
    It allows a schema (that inherits from :ref:`BaseSchema`) to passed in to use to validate and build.
    However, it assumes the ``AutograderConfigurationSchema`` will be used.

    This allows loading from currently only toml files, but is very easy to expand to different file formats if needed.

    ``.build`` should always be the last thing called.

    In the future, configuration will be allowed with this builder, but I would need to see the use case.
    """
    DEFAULT_CONFIG_FILE = "./config.toml"

    def __init__(self, configSchema: BaseSchema[T] = AutograderConfigurationSchema()):
        self.schema: BaseSchema[T] = configSchema
        self.data: Dict = {}

    def fromTOML(self: Builder, file=DEFAULT_CONFIG_FILE, merge=True) -> Builder:
        """
        Attempt to load the autograder config from the TOML config file.
        This file is assumed to be located in the same directory as the actual test cases
        """
        with open(file, 'rb') as rb:
            self.data = load(rb)

        return self

    # Really easy to add support for other file formats.
    # YAML or JSON would work as well

    @staticmethod
    def _createKeyIfDoesntExist(source: Dict[str, Any], key: str):
        if key in source:
            return

        source[key] = {}

    def setStudentSubmissionDirectory(self: Builder, studentSubmissionDirectory: OptionalType[str]) -> Builder:
        if studentSubmissionDirectory is None:
            return self

        self._createKeyIfDoesntExist(self.data, "config")

        self.data["config"]["student_submission_directory"] = studentSubmissionDirectory

        return self

    def setTestDirectory(self: Builder, testDirectory: str) -> Builder:
        self._createKeyIfDoesntExist(self.data, "config")

        self.data["config"]["test_directory"] = testDirectory

        return self

    def setAutograderRoot(self: Builder, autograderDirectory: str) -> Builder:
        self.data["autograder_root"] = autograderDirectory

        return self

    def build(self) -> T:
        self.data = self.schema.validate(self.data)
        return self.schema.build(self.data)


class AutograderConfigurationProvider:
    """
    AutograderConfigurationProvider
    ===============================

    This class allows access to the same config across the entire program.
    This is using a similar pattern to singletons, however, it's a bit better as it's a separate provider.
    """
    config: OptionalType[AutograderConfiguration] = None

    @classmethod
    def get(cls) -> AutograderConfiguration:
        if cls.config is None:
            raise AttributeError("Configuration has not been set!")

        return cls.config

    @classmethod
    def set(cls, config: AutograderConfiguration):
        if cls.config is not None:
            raise AttributeError("Configuration has already been set!")

        cls.config = config

    @classmethod
    def reset(cls):
        cls.config = None
